# coding: utf-8

# flake8: noqa

"""
    Meshcapade Me API

    Welcome to the age of Avatars, introducing the Meshcapade Me API. The no-friction avatar-creation platform for all your avatar needs. Create accurate digital doubles from any source of data in a unified 3D body format for every industry. Built on our <a href=\"https://meshcapade.com/SMPL\" target=\"_blank\">patented avatar</a> technology, our API allows you to create and edit avatars from images, videos, measurements, scans, and more. # Introduction The Meshcapade Me API is a RESTful API that allows you to create and edit avatars from images, measurements, scans, and more. All API replies adhere to the  <a href=\"https://jsonapi.org/format/\" target=\"_blank\">JSON:API</a> schema guidelines. Currently the API is in beta and is subject to change. Thus, not all ways to create avatars are available yet. We are working hard to add more ways to create avatars and will update this documentation accordingly. The API allows you to create avatars </br> - From <a href=\"#post-/avatars/create/from-images\" target=\"_blank\">images</a> </br> - From <a href=\"#post-/avatars/create/from-video\" target=\"_blank\">video</a> </br> - From <a href=\"#post-/avatars/create/from-scans\" target=\"_blank\">3D scans</a> </br> - From <a href=\"#post-/avatars/create/from-measurements\" target=\"_blank\">measurements</a> </br> # Quickstart To get started, sign up for a free account at <a href=\"https://me.meshcapade.com\" target=\"_blank\">me.meshcapade.com</a></br> We recommend using our <a href=\"https://www.postman.com/downloads/\" target=\"_blank\">Postman</a> collection to conveniently explore the API </br> <div style=\"margin-top: 16px;\"><a href=\"https://www.postman.com/cloudy-meadow-883625/workspace/meshcapade/overview\"><img src=\"https://run.pstmn.io/button.svg\" alt=\"Run in Postman\"></a></div></br>  # How-To <a href=\"https://medium.com/meshcapade/streamline-avatar-creation-with-meshcapade-me-api-from-one-image-to-an-accurate-avatar-in-seconds-b8ca4f15b9a8\" target=\"_blank\">Create an avatar from a single image (Medium)</a> </br> <a href=\"https://medium.com/meshcapade/measurements-meet-imagination-creating-accurate-3d-avatars-with-meshcapades-api-9a6ec5029793\" target=\"_blank\">Create an avatar from measurements (Medium)</a> # API Categories The API is organized into the following main categories: - Assets </br> Endpoint for listing assets of multiple types. - Avatars </br> Endpoints for listing, downloading, and deleting avatars. - Mesh </br> Endpoints for listing, downloading, and deleting exported meshes. - Images </br> Endpoints for listing, uploading, and deleting images related to avatars. - Create Avatar from images </br> Endpoints to initiate and complete avatar creation from images. - Create Avatar from measurements </br> Endpoint to initiate avatar creation from body measurements. - Create Avatar from scans </br> Endpoints to initiate and complete avatar creation from 3d body scans. - Create Avatar from betas </br> Endpoint to create an avatar from SMPL based beta shape parameters. # Error codes When something goes wrong, the API replies with an additional error code  - `asset_not_found` The requested asset either does not exist, or is not owned by the user (404)  </br> - `too_many_images` The image limit that can be uploaded for avatars from images as been exceeded. (400) </br> - `already_started` A process that already has been requested cannot be started again. (400) </br> - `no_images`  [POST /avatars/create/from-images](#post-/avatars/create/from-images) can only be started with at least one image uploaded  (400) </br> - `inputs_not_ready` Running `/avatars/create/xxxx` endpoint require the inputs to be uploaded (400) </br> - `uuid_invalid_format` Asset ID is in a non-UUID format (400) </br> - ` missing_parameters` Not all required parameters have been supplied for the request (400) </br> - `unauthorized` Trying to access an asset the user does not own, or endpoints that the user is not authorized to call (400) </br> - `too_many_builds` User processing rate has been exceeded. Only one computation heavy process can be started at a time (429) </br> - `asset_not_ready` Cannot call request on an asset which is not in a READY state (400) </br>  # Integration When integrating with our API, you may encounter Cross-Origin Resource Sharing (CORS) issues during deployment. </br> For security reasons, our API does not support direct communication from your frontend. </br> Instead, we recommend that you connect to api.meshcapade through your own backend server. </br> This approach not only mitigates CORS-related challenges but also enhances the overall security of your application. </br> By handling API requests server-side, you can ensure smoother and safer integration. </br> If you encounter any issues during integration, please reach out to us at support@meshcapade.com </br>  # noqa: E501

    The version of the OpenAPI document: v1.20
    Contact: support@meshcapade.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


__version__ = "1.0.0"

# import apis into sdk package
from openapi_client.api.assets_api import AssetsApi
from openapi_client.api.avatars_api import AvatarsApi
from openapi_client.api.blend_motions_api import BlendMotionsApi
from openapi_client.api.create_avatar_from_images_api import CreateAvatarFromImagesApi
from openapi_client.api.create_avatar_from_scans_api import CreateAvatarFromScansApi
from openapi_client.api.create_avatar_from_video_api import CreateAvatarFromVideoApi
from openapi_client.api.create_avatar_from_measurements_api import CreateAvatarFromMeasurementsApi
from openapi_client.api.create_avatars_from_betas_api import CreateAvatarsFromBetasApi
from openapi_client.api.create_scene_from_video_api import CreateSceneFromVideoApi
from openapi_client.api.credit_bundles_api import CreditBundlesApi
from openapi_client.api.helpers_api import HelpersApi
from openapi_client.api.images_api import ImagesApi
from openapi_client.api.info_api import InfoApi
from openapi_client.api.measure_avatar_api import MeasureAvatarApi
from openapi_client.api.mesh_api import MeshApi
from openapi_client.api.motions_api import MotionsApi
from openapi_client.api.search_motions_api import SearchMotionsApi
from openapi_client.api.subscription_api import SubscriptionApi
from openapi_client.api.user_api import UserApi
from openapi_client.api.videos_api import VideosApi

# import ApiClient
from openapi_client.api_response import ApiResponse
from openapi_client.api_client import ApiClient
from openapi_client.configuration import Configuration
from openapi_client.exceptions import OpenApiException
from openapi_client.exceptions import ApiTypeError
from openapi_client.exceptions import ApiValueError
from openapi_client.exceptions import ApiKeyError
from openapi_client.exceptions import ApiAttributeError
from openapi_client.exceptions import ApiException

# import models into sdk package
from openapi_client.models.docschemas_doc_afi_inputs import DocschemasDocAFIInputs
from openapi_client.models.docschemas_doc_afs_inputs import DocschemasDocAFSInputs
from openapi_client.models.docschemas_doc_afv_inputs import DocschemasDocAFVInputs
from openapi_client.models.docschemas_doc_asset_response import DocschemasDocAssetResponse
from openapi_client.models.docschemas_doc_avatar_attributes import DocschemasDocAvatarAttributes
from openapi_client.models.docschemas_doc_avatar_link import DocschemasDocAvatarLink
from openapi_client.models.docschemas_doc_avatar_metadata import DocschemasDocAvatarMetadata
from openapi_client.models.docschemas_doc_avatar_relationships import DocschemasDocAvatarRelationships
from openapi_client.models.docschemas_doc_avatar_relationships_exported_mesh import DocschemasDocAvatarRelationshipsExportedMesh
from openapi_client.models.docschemas_doc_avatar_relationships_exported_mesh_data_inner import DocschemasDocAvatarRelationshipsExportedMeshDataInner
from openapi_client.models.docschemas_doc_base_jsonapi_response import DocschemasDocBaseJSONAPIResponse
from openapi_client.models.docschemas_doc_base_mesh_no_includes_response import DocschemasDocBaseMeshNoIncludesResponse
from openapi_client.models.docschemas_doc_base_single_jsonapi_response import DocschemasDocBaseSingleJSONAPIResponse
from openapi_client.models.docschemas_doc_body_shape import DocschemasDocBodyShape
from openapi_client.models.docschemas_doc_build_attributes import DocschemasDocBuildAttributes
from openapi_client.models.docschemas_doc_build_jsonapi_response import DocschemasDocBuildJSONAPIResponse
from openapi_client.models.docschemas_doc_build_link import DocschemasDocBuildLink
from openapi_client.models.docschemas_doc_build_response import DocschemasDocBuildResponse
from openapi_client.models.docschemas_doc_create_from_smpl_request import DocschemasDocCreateFromSMPLRequest
from openapi_client.models.docschemas_doc_credit_bundle_attributes import DocschemasDocCreditBundleAttributes
from openapi_client.models.docschemas_doc_credit_bundle_response import DocschemasDocCreditBundleResponse
from openapi_client.models.docschemas_doc_credit_bundles_api_response import DocschemasDocCreditBundlesAPIResponse
from openapi_client.models.docschemas_doc_error_response import DocschemasDocErrorResponse
from openapi_client.models.docschemas_doc_export_inputs import DocschemasDocExportInputs
from openapi_client.models.docschemas_doc_mesh_attributes import DocschemasDocMeshAttributes
from openapi_client.models.docschemas_doc_mesh_link import DocschemasDocMeshLink
from openapi_client.models.docschemas_doc_mesh_measurements import DocschemasDocMeshMeasurements
from openapi_client.models.docschemas_doc_mesh_metadata import DocschemasDocMeshMetadata
from openapi_client.models.docschemas_doc_mesh_no_relation_response import DocschemasDocMeshNoRelationResponse
from openapi_client.models.docschemas_doc_mesh_response import DocschemasDocMeshResponse
from openapi_client.models.docschemas_doc_motion import DocschemasDocMotion
from openapi_client.models.docschemas_doc_motion_options import DocschemasDocMotionOptions
from openapi_client.models.docschemas_doc_motion_result import DocschemasDocMotionResult
from openapi_client.models.docschemas_doc_subscription_attributes import DocschemasDocSubscriptionAttributes
from openapi_client.models.docschemas_doc_user_attributes import DocschemasDocUserAttributes
from openapi_client.models.docschemas_doc_user_jsonapi_response import DocschemasDocUserJSONAPIResponse
from openapi_client.models.docschemas_doc_user_response import DocschemasDocUserResponse
from openapi_client.models.enums_gender import EnumsGender
from openapi_client.models.enums_model_version import EnumsModelVersion
from openapi_client.models.mesherr_mesh_err import MesherrMeshErr
from openapi_client.models.postgresql_asset_state import PostgresqlAssetState
from openapi_client.models.postgresql_asset_type import PostgresqlAssetType
from openapi_client.models.postgresql_build_method import PostgresqlBuildMethod
from openapi_client.models.postgresql_build_state import PostgresqlBuildState
from openapi_client.models.postgresql_file_source import PostgresqlFileSource
from openapi_client.models.schemas_betas_avatar_request import SchemasBetasAvatarRequest
from openapi_client.models.schemas_image_tag_links import SchemasImageTagLinks
from openapi_client.models.schemas_measurement_avatar_request import SchemasMeasurementAvatarRequest
from openapi_client.models.schemas_motion_blend_motion import SchemasMotionBlendMotion
from openapi_client.models.schemas_motion_blend_request import SchemasMotionBlendRequest
from openapi_client.models.schemas_url_response import SchemasURLResponse
from openapi_client.models.services_search_motions_options import ServicesSearchMotionsOptions



from keycloak import KeycloakOpenID, exceptions
import click

from .logger import log
from .helpers import State

def validate_access_token(keycloak_openid: KeycloakOpenID, keycloak_tokens: dict[str,str]) -> str:
    """Validate the access token on keyloak"""
    if keycloak_tokens:
        token: str = keycloak_tokens["access_token"] if "access_token" in keycloak_tokens else ""
        try:
            keycloak_openid.userinfo(token)
            log.debug("Used saved access token")
            return token
        except exceptions.KeycloakAuthenticationError:
            pass
    return ""

def get_refreshed_tokens(keycloak_openid: KeycloakOpenID, keycloak_tokens: dict[str,str]) -> dict[str,str]:
    """Use refresh token to get a new access token from keycloak"""
    if keycloak_tokens:
        refresh_token: str = keycloak_tokens["refresh_token"] if "refresh_token" in keycloak_tokens else ""
        try:
            refreshed_tokens: dict[str,str] = keycloak_openid.refresh_token(refresh_token)
            log.debug("Used refreshed access token")
            return refreshed_tokens
        except exceptions.KeycloakPostError:
            pass
    return {}

def generate_keycloak_tokens(keycloak_openid: KeycloakOpenID, username: str, password: str) -> dict[str,str]:
    """Get new keycloak access token with username and password"""
    keycloak_tokens: dict[str,str] = keycloak_openid.token(username, password)
    log.debug("Used generated access token")
    return keycloak_tokens

def authenticate(keycloak_openid: KeycloakOpenID, state: State, username: str, password: str) -> str:
    """Get keycloak access token by retrieving saved ones, refreshing it or requesting a new one with username and password"""
    # load state: get current user and saved auth tokens if there are any
    user, tokens = state.get_user_and_tokens(username)
    log.debug("User: " + user) if user is not None else log.debug("No active user")
    # use access token saved in state file if it is valid
    saved_access_token: str = validate_access_token(keycloak_openid, tokens)
    if saved_access_token != "":
        state.update(tokens, user)
        return saved_access_token
    # else: use refresh token saved in state file to refresh access token if it is valid, save to state file
    refreshed_tokens: dict[str,str] = get_refreshed_tokens(keycloak_openid, tokens)
    if refreshed_tokens != {}:
        state.update(refreshed_tokens, user)
        return refreshed_tokens["access_token"]
    # else: generate new access and refresh tokens using username and password, save to state file
    if user is None:
        user = click.prompt('Username', type=str)
    if password is None:
        password = click.prompt('Password', type=str, hide_input=True)
    generated_tokens: dict[str,str] = generate_keycloak_tokens(keycloak_openid, user, password)
    state.update(generated_tokens, user)
    return generated_tokens["access_token"]


import os
import click

from .logger import log
import openapi_client as client
from openapi_client.models.postgresql_build_method import PostgresqlBuildMethod

from .avatars import from_video, from_images, from_scans, export
from .helpers import LocalUploader, download_avatar, get_timestamp

def validate_batch_input_dir(ctx, param, value) -> bool:
    """
    Validate the input directory for batch create.
    - Check if the directory has more than one file
    """
    if not os.path.isdir(value):
        raise click.BadParameter("Input directory does not exist")
    
    if len(os.listdir(value)) < 1:
        raise click.BadParameter("Input directory does not have enough files")

    return value

def handle_batch_processing(api_client, batch_options_js, input_dir: str, output_dir: str):
    """
    Create batch of avatars from files.
    """
    log.info(f"Creating batch of avatars from files in {input_dir}")
    input_file_path = os.path.join(input_dir, "input.mp4")

    missing_files_handle = open(os.path.join(output_dir, "missing_files.txt"), "a")
    errored_processes = []

    # Iterate through batch_options_js["data"]["input_files"]
    # Read out "filename" and look for the file in the input directory
    # If the file is found, create the avatar
    for file in batch_options_js["data"]["input_files"]:
        file_name = file["filename"]
        input_file_path = os.path.join(input_dir, file_name)

        # Check if the file exists
        if not os.path.exists(input_file_path):
            log.warn(f"File {input_file_path} does not exist")
            missing_files_handle.write(f"{input_file_path}\n")
            continue

        # Create avatar
        avatar_id = get_processor(batch_options_js["mode"], input_file_path, file["parameters"], api_client, batch_options_js["timeout"])

        # TODO(dheid)
        # Make download and export configurable to be also running in parallel
        avatar_client = client.AvatarsApi(api_client)
        if file.get("export_parameters"):
            format = file["export_parameters"]["format"]
            pose = file["export_parameters"].get("pose")
            download_url = export(avatar_id, format, file["export_parameters"]["pose"], file["export_parameters"]["animation"], "default", avatar_client, batch_options_js["timeout"])
            out_filename = os.path.join(output_dir, f"{get_timestamp()}_{file_name}.{format.lower()}")
            download_avatar(out_filename, download_url)


    if errored_processes:
        log.info(f"Writing list of errored processes to {output_dir}")
        with open(os.path.join(output_dir, "errored_processes.txt"), "w") as f:
            f.write("\n".join(errored_processes))

    missing_files_handle.close()


def get_processor(mode, input_file_path: str, params: dict, api_client, timeout) -> str:
    """
    Get the function to create avatar based on mode.
    """
    avatar_client = client.AvatarsApi(api_client)
    uploader = LocalUploader()
    mode = mode.lower()

    params["avatarname"] = os.path.basename(input_file_path).split(".")[0]

    log.info(f"Creating avatar with mode {mode}")
    if mode.upper() == PostgresqlBuildMethod.FROM_VIDEO.value:
        from_video_client = client.CreateAvatarFromVideoApi(api_client)
        return from_video(params["gender"], params["avatarname"], input_file_path, from_video_client, avatar_client, uploader, timeout)
    else:
        raise NotImplementedError(f"Mode {mode} is not implemented")








import os
import json
from typing import Any, Optional

from keycloak import KeycloakOpenID
import openapi_client as client
import click
from openapi_client.models.postgresql_build_method import PostgresqlBuildMethod
from os import path, walk
from datetime import datetime
from .logger import log
from .auth import authenticate
from .helpers import (load_config, 
                      parse_betas, 
                      validate_export_parameter, 
                      get_export_parameters, 
                      download_avatar,
                      select_avatar_id,
                      get_timestamp,
                      get_measurements_dict,
                      Uploader, State,
                      select_motion_id)
from .motions import TMRMotion
from .avatars import from_betas, export, from_images, from_measurements, from_scans, from_video, from_smpl, blend_motions
from .batch_create import validate_batch_input_dir, handle_batch_processing
from .user import User, request_user_info
from .constants import CREDIT_COSTS
from functools import partial


CURRENT_DIR = path.dirname(path.abspath(__file__))
DEFAULT_CONFIG = path.join(CURRENT_DIR, "../configs/prod.toml")

@click.group()
@click.pass_context
@click.option('--config',
              type=click.Path(exists=True), default=os.environ.get("MCME_CONFIG_PATH", DEFAULT_CONFIG), help="Path to config file")
@click.option(
    "--username",
    default=lambda: os.environ.get("MCME_USERNAME")
)
@click.option(
    "--password",
    default=lambda: os.environ.get("MCME_PASSWORD")
)
def cli(ctx: click.Context, 
        username: str,
        password: str,
        config: str) -> None:
    """
    Command-line interface for the Meshcapade.me API.
    """
    ctx.ensure_object(dict)
    ctx.obj["config"] = load_config(config)
    ctx.obj["keycloak_tokens"] = os.path.expanduser(ctx.obj["config"]["cli_state"]["keycloak_tokens"])
    auth_config = ctx.obj["config"]["auth"]
    keycloak_openid: KeycloakOpenID = KeycloakOpenID(server_url=auth_config["server_url"],
                                    client_id=auth_config["client_id"],
                                    realm_name=auth_config["realm_name"])
    state = State(ctx.obj["keycloak_tokens"])
    ctx.obj["token"] = authenticate(keycloak_openid, state, username, password)
    # construct api client
    configuration = client.Configuration(
        host = ctx.obj["config"]["api"]["host"]
    )
    configuration.access_token =  ctx.obj["token"]
    # Enter a context with an instance of the API client
    ctx.obj["api_client"] = client.ApiClient(configuration)
    

@cli.result_callback()
@click.pass_context
def close_api_client(
    ctx: click.Context,
    result: Any,
    **kwargs):
    ctx.obj["api_client"].close()

@cli.group()
@click.pass_context
@click.option(
    "--download-format",
    type=click.Choice(["OBJ", "FBX"], case_sensitive=False),
    is_eager=True,
    help="Format for downloading avatar"
)
@click.option(
    "--pose",
    type=click.Choice(["T", "A", "I", "SCAN"], case_sensitive=False),
    callback=validate_export_parameter,
    help="Pose the downloaded avatar should be in. SCAN is not applicable for avatars created from betas or measurements since it corresponds to a captured pose or motion."
)
@click.option(
    "--animation",
    type=click.Choice(["a-salsa"], case_sensitive=False),
    callback=validate_export_parameter,
    help="Animation for the downloaded avatar"
)
@click.option(
    "--compatibility-mode",
    type=click.Choice(["DEFAULT", "OPTITEX", "UNREAL"], case_sensitive=False),
    callback=validate_export_parameter,
    help="Adjust output for compatibility with selected software."
)
@click.option(
    "--out-file",
    type=click.Path(dir_okay=False),
    callback=validate_export_parameter,
    help="File to save created avatar mesh to"
)
def create(
    ctx: click.Context, 
    download_format: str, 
    pose: str, 
    animation: str, 
    compatibility_mode: str, 
    out_file: click.Path
    ) -> None:
    """
    Create avatars. Please be aware that these commands cost credits.
    """
    # all create avatar operations need keycloak authentication
    ctx.obj["download"] = True if download_format is not None else False


@create.command(name="from-betas")
@click.pass_context
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              help="Gender of created avatar")
@click.option('--betas',
              type=click.UNPROCESSED, callback=partial(parse_betas, is_smplx=False), 
              help="Beta values. Supply like 0.1,0.2 or \"[0.1,0.2]\"")
@click.option('--name',
              type=str, 
              default="avatar_from_betas",
              help="Name of created avatar")
@click.option('--model-version',
              type=click.Choice(client.EnumsModelVersion.enum_values(), case_sensitive=False), 
              help="Model version")
def create_from_betas(
    ctx: click.Context,
    gender: Optional[client.EnumsGender], 
    betas: list[float], 
    name: str, 
    model_version: Optional[client.EnumsModelVersion]
    ) -> None:
    """Create avatar from betas."""
    # Create avatar from betas
    api_instance = client.CreateAvatarsFromBetasApi(ctx.obj["api_client"])
    asset_id = from_betas(gender, betas, name, model_version, api_instance)
    log.info(f"AssetID: {asset_id}")

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@create.command(name="from-measurements")
@click.pass_context
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              required=True,
              help="Gender of created avatar")
@click.option('--name',
              type=str, 
              default="avatar_from_measurements",
              help="Name of created avatar")
@click.option('--height',
              type=float,
              help="Height")
@click.option('--weight',
              type=float,
              help="Weight")
@click.option('--bust-girth',
              type=float,
              help="Bust girth")
@click.option('--ankle-girth',
              type=float,
              help="Ankle girth")
@click.option('--thigh-girth',
              type=float,
              help="Thigh girth")
@click.option('--waist-girth',
              type=float,
              help="Waist girth")
@click.option('--armscye-girth',
              type=float,
              help="Armscye girth")
@click.option('--top-hip-girth',
              type=float,
              help="Top hip girth")
@click.option('--neck-base-girth',
              type=float,
              help="Neck base girth")
@click.option('--shoulder-length',
              type=float,
              help="Shoulder length")
@click.option('--lower-arm-length',
              type=float,
              help="Lower arm length")
@click.option('--upper-arm-length',
              type=float,
              help="Upper arm length")
@click.option('--inside-leg-height',
              type=float,
              help="Inside leg height")
@click.option('--model-version',
              type=click.Choice(client.EnumsModelVersion.enum_values(), case_sensitive=False), 
              help="Model version")
def create_from_measurements(
    ctx: click.Context,
    gender: Optional[client.EnumsGender],
    name: str,
    height,
    weight,
    bust_girth,
    ankle_girth,
    thigh_girth,
    waist_girth,
    armscye_girth,
    top_hip_girth,
    neck_base_girth,
    shoulder_length,
    lower_arm_length,
    upper_arm_length,
    inside_leg_height,
    model_version: Optional[client.EnumsModelVersion]
    ) -> None:
    """Create avatar from measurements."""
    # Create avatar from measurements
    measurements = get_measurements_dict(
    height,
    weight,
    bust_girth,
    ankle_girth,
    thigh_girth,
    waist_girth,
    armscye_girth,
    top_hip_girth,
    neck_base_girth,
    shoulder_length,
    lower_arm_length,
    upper_arm_length,
    inside_leg_height
    )
    api_instance_from_measurements = client.CreateAvatarFromMeasurementsApi(ctx.obj["api_client"])
    api_instance_avatars = client.AvatarsApi(ctx.obj["api_client"])
    timeout = ctx.obj["config"]["cli"]["timeout"]
    asset_id = from_measurements(gender, name, measurements, model_version, api_instance_from_measurements, api_instance_avatars, timeout)

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)

@create.command(name="from-images")
@click.pass_context
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              help="Gender of created avatar")
@click.option('--name',
              type=str, 
              default="avatar_from_images",
              help="Name of created avatar")
@click.option('--input',
              type=click.Path(dir_okay=False, exists=True),
              help="Path to input image")
@click.option('--height',
              type=int,
              help="Height of the person in the image")
@click.option('--weight',
              type=int,
              help="Weight of the person in the image")
@click.option('--image-mode',
              type=click.Choice(["AFI", "BEDLAM_CLIFF"], case_sensitive=False),
              default="AFI",
              help="Mode for avatar creation")
def create_from_images(
    ctx: click.Context,
    gender: Optional[client.EnumsGender],
    name: str,
    input: str,
    height: int,
    weight: int,
    image_mode: str
    ) -> None:
    """Create avatar from images."""
    api_instance_images = client.CreateAvatarFromImagesApi(ctx.obj["api_client"])
    api_instance_avatars = client.AvatarsApi(ctx.obj["api_client"])
    uploader = Uploader()
    timeout = ctx.obj["config"]["cli"]["timeout"]
    asset_id = from_images(gender, name, input, height, weight, image_mode, api_instance_images, api_instance_avatars, uploader, timeout)

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@create.command(name="from-scans")
@click.pass_context
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              help="Gender of created avatar")
@click.option('--name',
              type=str, 
              default="avatar_from_scans",
              help="Name of created avatar")
@click.option('--input',
              type=click.Path(dir_okay=False, exists=True),
              help="Path to input image")
@click.option('--init-pose',
              type=str,
              help="Pose for initialization")
@click.option('--up-axis',
              type=str,
              help="Up axis")
@click.option('--look-axis',
              type=str,
              help="Look axis")
@click.option('--input-units',
              type=str, 
              help="Input units of scan")
def create_from_scans(
    ctx: click.Context,
    gender: Optional[client.EnumsGender],
    name: str,
    input: str,
    init_pose: str,
    up_axis: str,
    look_axis: str,
    input_units: str
    ) -> None:
    """Create avatar from images."""
    api_instance_scan = client.CreateAvatarFromScansApi(ctx.obj["api_client"])
    api_instance_avatars = client.AvatarsApi(ctx.obj["api_client"])
    uploader = Uploader()
    timeout = ctx.obj["config"]["cli"]["timeout"]
    asset_id = from_scans(gender, name, input, init_pose, up_axis, look_axis, input_units, api_instance_scan, api_instance_avatars, uploader, timeout)

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@create.command(name="from-video")
@click.pass_context
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              help="Gender of created avatar")
@click.option('--name',
              type=str, 
              default="avatar_from_images",
              help="Name of created avatar")
@click.option('--input',
              type=click.Path(dir_okay=False, exists=True),
              help="Path to input video")
def create_from_video(
    ctx: click.Context,
    gender: Optional[client.EnumsGender],
    name: str,
    input: str
    ) -> None:
    """Create avatar from a single video."""
    api_instance_images = client.CreateAvatarFromVideoApi(ctx.obj["api_client"])
    api_instance_avatars = client.AvatarsApi(ctx.obj["api_client"])
    uploader = Uploader()
    timeout = ctx.obj["config"]["cli"]["timeout"]
    asset_id = from_video(gender, name, input, api_instance_images, api_instance_avatars, uploader, timeout)

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@create.command(name="from-text")
@click.pass_context
@click.option('--prompt',
              type=str, 
              required=True,
              help="Text prompt describing desired motion")
@click.option('--name',
              type=str,
              help="Name of created avatar")
def create_from_text(
    ctx: click.Context,
    prompt: str,
    name: str
    ) -> None:
    """Create avatar with motion from text prompt."""
    api_instance_motions = client.SearchMotionsApi(ctx.obj["api_client"])
    api_instance_avatars = client.AvatarsApi(ctx.obj["api_client"])
    timeout = ctx.obj["config"]["cli"]["timeout"]

    if name is None:
        name = prompt.replace(" ", "_")

    motion = TMRMotion(prompt)

    # Search for motion by prompt and save temporary smpl file
    motion.find_motion(api_instance_motions)

    trimmed_motion = motion.trim()

    # Use found and trimmed motion .smpl file to create avatar
    uploader = Uploader()
    asset_id = from_smpl(trimmed_motion, api_instance_avatars, name, uploader, timeout)

    # Delete temporary motion .smpl file
    motion.cleanup()

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@create.command(name="blend-motions")
@click.pass_context
@click.option('--source-avatar-id',
              type=str,
              help="A clone of this avatar will be used to attach the resulting motion to")
@click.option("--motion-id-1", 
              type=str,
              help="First motion to use for blending")
@click.option("--motion-id-2", 
              type=str,
              help="Second motion to use for blending")
@click.option('--shape-parameters',
              type=click.UNPROCESSED, callback=partial(parse_betas, is_smplx=True), 
              help="If source-avatar-id is not specified, these SMPLX shape parameters will be used to create avatar from. Supply like 0.1,0.2 or \"[0.1,0.2]\"")
@click.option('--gender',
              type=click.Choice(client.EnumsGender.enum_values(), case_sensitive=False),
              help="If source-avatar-id is not specified, this gender will be used for created avatar.")
@click.option('--name',
              type=str, 
              default="avatar_from_blend_motion",
              help="Name of created avatar")
def create_blend_motions(
    ctx: click.Context,
    source_avatar_id: str,
    motion_id_1: str,
    motion_id_2: str,
    shape_parameters: Optional[list[float]],
    gender: Optional[client.EnumsGender], 
    name: str
    ) -> None:
    """Create avatar with blended motion attached."""
    api_instance = client.BlendMotionsApi(ctx.obj["api_client"])
    motions_api_instance = client.MotionsApi(ctx.obj["api_client"])
    avatar_api_instance = client.AvatarsApi(ctx.obj["api_client"])
    if shape_parameters == []:
        shape_parameters = None
    timeout = ctx.obj["config"]["cli"]["timeout"]

    if motion_id_1 is None:
        motion_id_1 = select_motion_id(motions_api_instance.list_motions, 10, "first ")
    if motion_id_2 is None:
        motion_id_2 = select_motion_id(motions_api_instance.list_motions, 10, "second ")

    asset_id = blend_motions(source_avatar_id, motion_id_1, motion_id_2, shape_parameters, gender, name, api_instance, avatar_api_instance, timeout)
    log.info(f"AssetID: {asset_id}")

    # Exit here if avatar should not be downloaded
    if ctx.obj["download"]:
        # Get download parameters from parent context
        params = get_export_parameters(ctx)
        if params.download_format != "FBX" or params.pose != "SCAN":
            raise click.BadArgumentUsage("Dowwnload format has to be FBX and pose has to be SCAN to attach blended motion to the resulting avatar.")
        # Export avatar
        ctx.invoke(export_and_download_avatar, 
                   format=params.download_format,
                   pose=params.pose,
                   animation=params.animation,
                   compatibility_mode=params.compatibility_mode,
                   out_file=params.out_file,
                   asset_id=asset_id)


@cli.command(name="batch-create")
@click.pass_context
@click.option('--options',
              type=click.Path(dir_okay=False),
              required=True,
              help="Path to options file. This file should contain a list of dictionaries, each containing the parameters for creating the avatars.")
@click.option('--input-dir',
              type=click.Path(dir_okay=True),
              required=True,
              callback=validate_batch_input_dir,
              help="Path to directory containing input files.")
@click.option('--output-dir',
              type=click.Path(dir_okay=True),
              required=True,
              help="Path to directory where to save the results to.")
def batch_create(
    ctx: click.Context,
    options: click.Path,
    input_dir: click.Path,
    output_dir: click.Path,
    ) -> None:
    """
    Create multiple avatars in batch mode. Please be aware that these commands cost credits.
    """
    # Read out batch options file
    batch_options = {}
    with open(options, "r") as f:
        batch_options = json.load(f)

    # Match batch mode to the correct function
    if batch_options.get("mode") is None:
        raise click.ClickException("Batch mode not specified")

    # Validate the process mode
    if PostgresqlBuildMethod(batch_options.get("mode").upper()) is None:
        raise click.ClickException("Invalid process mode")

    batch_options["timeout"] = ctx.obj["config"]["cli"]["timeout"]

    
    api_instance_user = client.UserApi(ctx.obj["api_client"])

    user = request_user_info(api_instance_user)

    # Read from constants the price of the batch creation for each file
    # TODO(dheid): Get the actual price for one alignment from the API and use that when that endpoint is available
    if user.credits < len(batch_options["data"]["input_files"]) * CREDIT_COSTS[batch_options.get("mode").upper()]:
        log.info(f"Insufficient credits for batch creation. Credits: {user.credits}. Still want to proceed?")
        if not click.confirm("Do you want to proceed?"):
            raise click.ClickException("Insufficient credits for batch creation")

    handle_batch_processing(ctx.obj["api_client"], batch_options, input_dir, output_dir)
    

@cli.command(name="download")
@click.pass_context
@click.option(
    "--format",
    type=click.Choice(["OBJ", "FBX"], case_sensitive=False),
    default="OBJ",
    help="Format for downloading avatar"
)
@click.option(
    "--pose",
    type=click.Choice(["T", "A", "SCAN", "U", "I", "W"], case_sensitive=False),
    default = "A",
    help="Pose the downloaded avatar should be in"
)
@click.option(
    "--animation",
    type=click.Choice(["a-salsa"], case_sensitive=False),
    help="Animation for the downloaded avatar"
)
@click.option(
    "--compatibility-mode",
    type=click.Choice(["DEFAULT", "OPTITEX", "UNREAL"], case_sensitive=False),
    help="Compatibility mode"
)
@click.option(
    "--out-file",
    type=click.Path(dir_okay=False),
    help="File to save created avatar mesh to"
)
@click.option(
    "--asset-id",
    type=str,
    help="Asset id of avatar to be downloaded"
)
@click.option(
    "--show-max-avatars",
    type=int,
    default=10,
    help="Maximum number of created avatars to show (most recent ones are shown first)"
)
def export_and_download_avatar(
    ctx: click.Context,
    format: str, 
    pose: str, 
    animation: str, 
    compatibility_mode: str, 
    out_file: click.Path,
    asset_id: str,
    show_max_avatars: int
    ) -> None:
    """
    Export avatar using asset id.
    """
    api_instance = client.AvatarsApi(ctx.obj["api_client"])
    name = None
    if asset_id is None:
        asset_id, name = select_avatar_id(api_instance.list_avatars, show_max_avatars)

    # Export avatar
    timeout = ctx.obj["config"]["cli"]["timeout"]
    download_url = export(asset_id, format, pose, animation, compatibility_mode, api_instance, timeout)
    out_filename = str(out_file) if out_file is not None else \
        f"{get_timestamp()}_{name}.{format.lower()}" if (name is not None and name != "") else \
        f"{get_timestamp()}_{asset_id}.{format.lower()}"
    download_avatar(out_filename, download_url)
    log.info(f"Downloaded avatar to {out_filename}")

@cli.command()
@click.pass_context
def info(ctx: click.Context) -> None:
    """Show API info."""
    # Create an instance of the API class
    api_instance = client.InfoApi(ctx.obj["api_client"])

    try:
        # Show API info
        api_response: str = api_instance.info()
        log.info(api_response)
    except Exception as e:
        log.info("Exception when calling InfoApi->info: %s\n" % e)

@cli.command(name="user-info")
@click.pass_context
def user_info(ctx: click.Context) -> None:
    """Show username and available credits."""
    api_instance_user = client.UserApi(ctx.obj["api_client"])

    user = request_user_info(api_instance_user)

    log.info(f"Username: {user.email}")
    log.info(f"Credits: {user.credits}")



import logging

log = logging.getLogger("mcme-cli")
log.setLevel(logging.INFO)
console = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
console.setFormatter(formatter)
log.addHandler(console)
log.propagate = False


from openapi_client import PostgresqlAssetState, SchemasBetasAvatarRequest, SchemasMeasurementAvatarRequest, DocschemasDocAFIInputs, DocschemasDocExportInputs, DocschemasDocAFVInputs
from keycloak import exceptions
import uuid

class SchemasURLResponse:
    def __init__(self):
        self.path = "example_url.com"

class AvatarAttributes:
    def __init__(self, state: PostgresqlAssetState, url: SchemasURLResponse):
        self.state = state
        self.url = url

class AssetResponse:
    def __init__(self, avatar_attributes: AvatarAttributes, id: str):
        self.attributes = avatar_attributes
        self.id = id

class ApiResponse:
    def __init__(self, state: PostgresqlAssetState):
        state = state
        url = SchemasURLResponse()
        attributes = AvatarAttributes(state=state, url=url)
        id = uuid.uuid1()
        self.data = AssetResponse(avatar_attributes=attributes, id=str(id))  


class CreateAvatarsFromBetasApi:
    def create_avatar_from_betas(self, request: SchemasBetasAvatarRequest):
        api_response = ApiResponse(PostgresqlAssetState("READY"))
        return api_response


class CreateAvatarFromMeasurementsApi:
    def avatar_from_measurements(self, request: SchemasMeasurementAvatarRequest):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_PROCESSING"))
        return api_response
    
class AvatarsApi:
    def __init__(self):
        self.succeed_on_try = 1
        self.counter = 0

    def describe_avatar(self, asset_id: str):
        self.counter += 1
        if self.counter >= self.succeed_on_try: 
            api_response = ApiResponse(PostgresqlAssetState("READY"))
        else:
            api_response = ApiResponse(PostgresqlAssetState("PROCESSING"))
        return api_response
    
    def export_avatar(self, asset_id: uuid.UUID, input: DocschemasDocExportInputs):  
        self.counter += 1
        if self.counter >= self.succeed_on_try: 
            api_response = ApiResponse(PostgresqlAssetState("READY"))
        else:
            api_response = ApiResponse(PostgresqlAssetState("PROCESSING"))
        return api_response
    
    def set_number_of_tries(self, succeed_on: int):
        self.succeed_on_try = succeed_on

class CreateAvatarFromImagesApi:
    def create_avatar_from_images(self):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_AFI_INPUT"))
        return api_response

    def upload_image_to_avatar(self, asset_id: uuid.UUID):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_UPLOAD"))
        return api_response
    
    def avatar_fit_to_images(self, asset_id: uuid.UUID, afi_inputs: DocschemasDocAFIInputs):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_PROCESSING"))
        return api_response 
    
class CreateAvatarFromScansApi:
    def create_avatar_from_scans(self):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_AFS_INPUT"))
        return api_response

    def upload_mesh_to_avatar(self, asset_id: uuid.UUID):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_UPLOAD"))
        return api_response
    
    def avatar_fit_to_scans(self, asset_id: uuid.UUID, afi_inputs: DocschemasDocAFIInputs):
        api_response = ApiResponse(PostgresqlAssetState("AWAITING_PROCESSING"))
        return api_response 


class MockUploader:
    def upload(self, file_to_upload: str, upload_url: str):
        pass

class KeycloakOpenID:
    def __init__(self):
        self.valid = True

    def userinfo(self, token: str):
        if self.valid and token == "valid_token":
            return
        raise exceptions.KeycloakAuthenticationError()

    def set_valid(self, valid: bool):
        self.valid = valid

    def refresh_token(self, refresh_token: str):
        if self.valid and refresh_token == "valid_token":
            return {"access_token": "valid_token"}
        raise exceptions.KeycloakPostError

    def token(self, username: str, password: str) -> dict[str,str]:
        if self.valid:
            return {"access_token": "valid_token"}
        return {}

class MockState:
    def __init__(self) -> None:
        """Initialize mocked state with valid tokens and user"""
        self.state_file: str = ""
        self.active_user: str = "test_user"
        self.auth_tokens: dict[str, dict[str, str]] = {"test_user": {"access_token": "valid_token", "refresh_token": "valid_token"}}
    
    def invalidate_token(self, user: str, key: str):
        self.auth_tokens[user][key] = "invalid_token"
    
    def get_user_and_tokens(self, username: str) -> tuple[str, dict[str,str]]:
        """Load state from a local file"""
        return username, self.auth_tokens[username]

    def update(self, keycloak_tokens: dict[str, str], username: str) -> None:
        pass



# coding: utf-8

"""
    Meshcapade Me API

    Welcome to the age of Avatars, introducing the Meshcapade Me API. The no-friction avatar-creation platform for all your avatar needs. Create accurate digital doubles from any source of data in a unified 3D body format for every industry. Built on our <a href=\"https://meshcapade.com/SMPL\" target=\"_blank\">patented avatar</a> technology, our API allows you to create and edit avatars from images, videos, measurements, scans, and more. # Introduction The Meshcapade Me API is a RESTful API that allows you to create and edit avatars from images, measurements, scans, and more. All API replies adhere to the  <a href=\"https://jsonapi.org/format/\" target=\"_blank\">JSON:API</a> schema guidelines. Currently the API is in beta and is subject to change. Thus, not all ways to create avatars are available yet. We are working hard to add more ways to create avatars and will update this documentation accordingly. The API allows you to create avatars </br> - From <a href=\"#post-/avatars/create/from-images\" target=\"_blank\">images</a> </br> - From <a href=\"#post-/avatars/create/from-video\" target=\"_blank\">video</a> </br> - From <a href=\"#post-/avatars/create/from-scans\" target=\"_blank\">3D scans</a> </br> - From <a href=\"#post-/avatars/create/from-measurements\" target=\"_blank\">measurements</a> </br> # Quickstart To get started, sign up for a free account at <a href=\"https://me.meshcapade.com\" target=\"_blank\">me.meshcapade.com</a></br> We recommend using our <a href=\"https://www.postman.com/downloads/\" target=\"_blank\">Postman</a> collection to conveniently explore the API </br> <div style=\"margin-top: 16px;\"><a href=\"https://www.postman.com/cloudy-meadow-883625/workspace/meshcapade/overview\"><img src=\"https://run.pstmn.io/button.svg\" alt=\"Run in Postman\"></a></div></br>  # How-To <a href=\"https://medium.com/meshcapade/streamline-avatar-creation-with-meshcapade-me-api-from-one-image-to-an-accurate-avatar-in-seconds-b8ca4f15b9a8\" target=\"_blank\">Create an avatar from a single image (Medium)</a> </br> <a href=\"https://medium.com/meshcapade/measurements-meet-imagination-creating-accurate-3d-avatars-with-meshcapades-api-9a6ec5029793\" target=\"_blank\">Create an avatar from measurements (Medium)</a> # API Categories The API is organized into the following main categories: - Assets </br> Endpoint for listing assets of multiple types. - Avatars </br> Endpoints for listing, downloading, and deleting avatars. - Mesh </br> Endpoints for listing, downloading, and deleting exported meshes. - Images </br> Endpoints for listing, uploading, and deleting images related to avatars. - Create Avatar from images </br> Endpoints to initiate and complete avatar creation from images. - Create Avatar from measurements </br> Endpoint to initiate avatar creation from body measurements. - Create Avatar from scans </br> Endpoints to initiate and complete avatar creation from 3d body scans. - Create Avatar from betas </br> Endpoint to create an avatar from SMPL based beta shape parameters. # Error codes When something goes wrong, the API replies with an additional error code  - `asset_not_found` The requested asset either does not exist, or is not owned by the user (404)  </br> - `too_many_images` The image limit that can be uploaded for avatars from images as been exceeded. (400) </br> - `already_started` A process that already has been requested cannot be started again. (400) </br> - `no_images`  [POST /avatars/create/from-images](#post-/avatars/create/from-images) can only be started with at least one image uploaded  (400) </br> - `inputs_not_ready` Running `/avatars/create/xxxx` endpoint require the inputs to be uploaded (400) </br> - `uuid_invalid_format` Asset ID is in a non-UUID format (400) </br> - ` missing_parameters` Not all required parameters have been supplied for the request (400) </br> - `unauthorized` Trying to access an asset the user does not own, or endpoints that the user is not authorized to call (400) </br> - `too_many_builds` User processing rate has been exceeded. Only one computation heavy process can be started at a time (429) </br> - `asset_not_ready` Cannot call request on an asset which is not in a READY state (400) </br>  # Integration When integrating with our API, you may encounter Cross-Origin Resource Sharing (CORS) issues during deployment. </br> For security reasons, our API does not support direct communication from your frontend. </br> Instead, we recommend that you connect to api.meshcapade through your own backend server. </br> This approach not only mitigates CORS-related challenges but also enhances the overall security of your application. </br> By handling API requests server-side, you can ensure smoother and safer integration. </br> If you encounter any issues during integration, please reach out to us at support@meshcapade.com </br>  # noqa: E501

    The version of the OpenAPI document: v1.20
    Contact: support@meshcapade.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import io
import json
import logging
import re
import ssl

from urllib.parse import urlencode, quote_plus
import urllib3

from openapi_client.exceptions import ApiException, UnauthorizedException, ForbiddenException, NotFoundException, ServiceException, ApiValueError, BadRequestException


logger = logging.getLogger(__name__)


class RESTResponse(io.IOBase):

    def __init__(self, resp):
        self.urllib3_response = resp
        self.status = resp.status
        self.reason = resp.reason
        self.data = resp.data

    def getheaders(self):
        """Returns a dictionary of the response headers."""
        return self.urllib3_response.headers

    def getheader(self, name, default=None):
        """Returns a given response header."""
        return self.urllib3_response.headers.get(name, default)


class RESTClientObject(object):

    def __init__(self, configuration, pools_size=4, maxsize=None):
        # urllib3.PoolManager will pass all kw parameters to connectionpool
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/poolmanager.py#L75  # noqa: E501
        # https://github.com/shazow/urllib3/blob/f9409436f83aeb79fbaf090181cd81b784f1b8ce/urllib3/connectionpool.py#L680  # noqa: E501
        # maxsize is the number of requests to host that are allowed in parallel  # noqa: E501
        # Custom SSL certificates and client certificates: http://urllib3.readthedocs.io/en/latest/advanced-usage.html  # noqa: E501

        # cert_reqs
        if configuration.verify_ssl:
            cert_reqs = ssl.CERT_REQUIRED
        else:
            cert_reqs = ssl.CERT_NONE

        addition_pool_args = {}
        if configuration.assert_hostname is not None:
            addition_pool_args['assert_hostname'] = configuration.assert_hostname  # noqa: E501

        if configuration.retries is not None:
            addition_pool_args['retries'] = configuration.retries

        if configuration.tls_server_name:
            addition_pool_args['server_hostname'] = configuration.tls_server_name


        if configuration.socket_options is not None:
            addition_pool_args['socket_options'] = configuration.socket_options

        if maxsize is None:
            if configuration.connection_pool_maxsize is not None:
                maxsize = configuration.connection_pool_maxsize
            else:
                maxsize = 4

        # https pool manager
        if configuration.proxy:
            self.pool_manager = urllib3.ProxyManager(
                num_pools=pools_size,
                maxsize=maxsize,
                cert_reqs=cert_reqs,
                ca_certs=configuration.ssl_ca_cert,
                cert_file=configuration.cert_file,
                key_file=configuration.key_file,
                proxy_url=configuration.proxy,
                proxy_headers=configuration.proxy_headers,
                **addition_pool_args
            )
        else:
            self.pool_manager = urllib3.PoolManager(
                num_pools=pools_size,
                maxsize=maxsize,
                cert_reqs=cert_reqs,
                ca_certs=configuration.ssl_ca_cert,
                cert_file=configuration.cert_file,
                key_file=configuration.key_file,
                **addition_pool_args
            )

    def request(self, method, url, query_params=None, headers=None,
                body=None, post_params=None, _preload_content=True,
                _request_timeout=None):
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param query_params: query parameters in the url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ['GET', 'HEAD', 'DELETE', 'POST', 'PUT',
                          'PATCH', 'OPTIONS']

        if post_params and body:
            raise ApiValueError(
                "body parameter cannot be used with post_params parameter."
            )

        post_params = post_params or {}
        headers = headers or {}
        # url already contains the URL query string
        # so reset query_params to empty dict
        query_params = {}

        timeout = None
        if _request_timeout:
            if isinstance(_request_timeout, (int,float)):  # noqa: E501,F821
                timeout = urllib3.Timeout(total=_request_timeout)
            elif (isinstance(_request_timeout, tuple) and
                  len(_request_timeout) == 2):
                timeout = urllib3.Timeout(
                    connect=_request_timeout[0], read=_request_timeout[1])

        try:
            # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
            if method in ['POST', 'PUT', 'PATCH', 'OPTIONS', 'DELETE']:

                # no content type provided or payload is json
                if not headers.get('Content-Type') or re.search('json', headers['Content-Type'], re.IGNORECASE):
                    request_body = None
                    if body is not None:
                        request_body = json.dumps(body)
                    r = self.pool_manager.request(
                        method, url,
                        body=request_body,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers)
                elif headers['Content-Type'] == 'application/x-www-form-urlencoded':  # noqa: E501
                    r = self.pool_manager.request(
                        method, url,
                        fields=post_params,
                        encode_multipart=False,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers)
                elif headers['Content-Type'] == 'multipart/form-data':
                    # must del headers['Content-Type'], or the correct
                    # Content-Type which generated by urllib3 will be
                    # overwritten.
                    del headers['Content-Type']
                    r = self.pool_manager.request(
                        method, url,
                        fields=post_params,
                        encode_multipart=True,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers)
                # Pass a `string` parameter directly in the body to support
                # other content types than Json when `body` argument is
                # provided in serialized form
                elif isinstance(body, str) or isinstance(body, bytes):
                    request_body = body
                    r = self.pool_manager.request(
                        method, url,
                        body=request_body,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers)
                else:
                    # Cannot generate the request from given parameters
                    msg = """Cannot prepare a request message for provided
                             arguments. Please check that your arguments match
                             declared content type."""
                    raise ApiException(status=0, reason=msg)
            # For `GET`, `HEAD`
            else:
                r = self.pool_manager.request(method, url,
                                              fields={},
                                              preload_content=_preload_content,
                                              timeout=timeout,
                                              headers=headers)
        except urllib3.exceptions.SSLError as e:
            msg = "{0}\n{1}".format(type(e).__name__, str(e))
            raise ApiException(status=0, reason=msg)

        if _preload_content:
            r = RESTResponse(r)

            # log response body
            logger.debug("response body: %s", r.data)

        if not 200 <= r.status <= 299:
            if r.status == 400:
                raise BadRequestException(http_resp=r)

            if r.status == 401:
                raise UnauthorizedException(http_resp=r)

            if r.status == 403:
                raise ForbiddenException(http_resp=r)

            if r.status == 404:
                raise NotFoundException(http_resp=r)

            if 500 <= r.status <= 599:
                raise ServiceException(http_resp=r)

            raise ApiException(http_resp=r)

        return r

    def get_request(self, url, headers=None, query_params=None, _preload_content=True,
            _request_timeout=None):
        return self.request("GET", url,
                            headers=headers,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            query_params=query_params)

    def head_request(self, url, headers=None, query_params=None, _preload_content=True,
             _request_timeout=None):
        return self.request("HEAD", url,
                            headers=headers,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            query_params=query_params)

    def options_request(self, url, headers=None, query_params=None, post_params=None,
                body=None, _preload_content=True, _request_timeout=None):
        return self.request("OPTIONS", url,
                            headers=headers,
                            query_params=query_params,
                            post_params=post_params,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            body=body)

    def delete_request(self, url, headers=None, query_params=None, body=None,
               _preload_content=True, _request_timeout=None):
        return self.request("DELETE", url,
                            headers=headers,
                            query_params=query_params,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            body=body)

    def post_request(self, url, headers=None, query_params=None, post_params=None,
             body=None, _preload_content=True, _request_timeout=None):
        return self.request("POST", url,
                            headers=headers,
                            query_params=query_params,
                            post_params=post_params,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            body=body)

    def put_request(self, url, headers=None, query_params=None, post_params=None,
            body=None, _preload_content=True, _request_timeout=None):
        return self.request("PUT", url,
                            headers=headers,
                            query_params=query_params,
                            post_params=post_params,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            body=body)

    def patch_request(self, url, headers=None, query_params=None, post_params=None,
              body=None, _preload_content=True, _request_timeout=None):
        return self.request("PATCH", url,
                            headers=headers,
                            query_params=query_params,
                            post_params=post_params,
                            _preload_content=_preload_content,
                            _request_timeout=_request_timeout,
                            body=body)



from typing import Callable
import requests
import tomli
import os
import json
import click
import time
from datetime import datetime
import re
from typing_extensions import TypedDict
import openapi_client as client
from openapi_client import DocschemasDocAssetResponse as AssetResponse
from openapi_client import PostgresqlAssetState as AssetState
from .logger import log
from .schemas import ExportParameters

MIME_JPG = "image/jpg"
MIME_PLAIN = "text/plain"
MIME_PNG = "image/png"
MIME_VIDEO = "video/mp4"
MIME_BINARY = "application/octet-stream"
MIME_TYPES = {"jpg": MIME_JPG, "jpeg": MIME_JPG, "png": MIME_PNG, "obj": MIME_PLAIN, "mp4": MIME_VIDEO, "smpl": MIME_BINARY}

def load_config(config: str) -> dict[str, str]:
    """Load config"""
    with open(config, mode="rb") as conf:
        return tomli.load(conf)

class State:
    def __init__(self, keycloak_file: str) -> None:
        """Load state from a local file"""
        self.state_file: str = keycloak_file
        self.active_user: str = ""
        self.auth_tokens: dict[str, dict[str, str]] = {}
        if not os.path.exists(self.state_file):
            return
        with open(self.state_file) as state:
            state_json = json.load(state)
            self.active_user = state_json.get("active_user")
            self.auth_tokens = state_json.get("keycloak_tokens")

    def get_user_and_tokens(self, username: str) -> tuple[str, dict[str,str]]:
        """Load state from a local file"""
        user = username if username is not None else self.active_user
        if user in self.auth_tokens:
            return user, self.auth_tokens[user]
        else:
            return user, {}

    def update(self, keycloak_tokens: dict[str, str], username: str) -> None:
        """Save the generated access tokens in a local file"""
        create_parent_dir_if_not_exists(self.state_file)
        self.active_user = username
        self.auth_tokens[username] = keycloak_tokens

        with open(self.state_file, "w") as state_file:
            json.dump(self.to_dict(), state_file)

    def to_dict(self) -> dict:
        return {"active_user": self.active_user, "keycloak_tokens": self.auth_tokens}


def create_parent_dir_if_not_exists(keycloak_file: str) -> None:
    parent_dir = os.path.dirname(keycloak_file)
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir, exist_ok=True)

def parse_betas(ctx: click.Context, param: click.Parameter, value: str, is_smplx: bool) -> list[float]:
    """Parse betas from command line"""
    shape_params = [float(x) for x in value.strip('[').strip(']').split(',')] if value is not None else []
    if is_smplx and value is not None and len(shape_params) != 10:
        raise click.BadArgumentUsage("If supplied, shape parameters have to have exactly 10 values.")
    return shape_params

def validate_export_parameter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    """Check export parameters for direct download after creation"""
    if ctx.params["download_format"] is None:
        if value is not None:
            raise click.BadOptionUsage(str(param.name), f"Please use option --download_format if you want to download avatar. Otherwise, please leave out option --{param.name}.", ctx=None)
    # Set pose default value if download is requested
    elif param.name == "pose" and value is None:
        return "A"
    return value

def get_export_parameters(ctx: click.Context) -> ExportParameters:
    """Get download parameters from parent context."""
    if ctx is None or ctx.parent is None:
        raise click.ClickException("Parent context empty")
    download_format = ctx.parent.params["download_format"]
    pose = ctx.parent.params["pose"]
    animation = ctx.parent.params["animation"]
    compatibility_mode = ctx.parent.params["compatibility_mode"]
    out_file = ctx.parent.params["out_file"]
    return ExportParameters(download_format, pose, animation, compatibility_mode, out_file)

def download_avatar(out_filename: str, download_url: str) -> None:
    """Download avatar using presigned aws s3 url."""
    with open(out_filename, 'wb') as file:
        try:
            stream = requests.get(download_url, stream=True, timeout=60)
            stream.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise click.ClickException(str(err)) from err
        for chunk in stream.iter_content(chunk_size=1024 * 1024):
            file.write(chunk)

def parse_asset(avatar: AssetResponse) -> tuple[str, dict]:
    """Parse asset from api response and return as dict with selected attributes"""
    if avatar.attributes is not None and \
        avatar.attributes.created_at is not None and \
            avatar.attributes.origin is not None and \
                avatar.attributes.state is not None:
        created_at = datetime.fromtimestamp(avatar.attributes.created_at).strftime("%Y/%m/%d %H:%M:%S")
        created_from_match = re.search("(?<=\\.).*", str(avatar.attributes.origin))
        created_from = created_from_match.group() if created_from_match is not None else ""
        name = avatar.attributes.name
        asset_id = avatar.id
        return avatar.attributes.state, {"assetID": asset_id, "name": name, "timestamp": created_at, "origin": created_from}
    return "", {}

def get_ready_assets(list_function: Callable, show_max_assets: int) -> list[dict]:
    """List assets that have state READY"""
    assets: list[dict] = []
    page = 1
    # iterate through pages and collect avatars with state ready until there are enough
    while len(assets) < show_max_assets:
        api_response = list_function(limit=show_max_assets*2, page=page)
        if api_response.data is None:
            raise click.ClickException("Response came back empty")
        if len(api_response.data) == 0:
            break
        for asset_response in api_response.data:
            state, asset = parse_asset(asset_response)
            if state == AssetState.READY:
                assets.append(asset)
            if len(assets) == show_max_assets:
                break
        page += 1
    return assets

def select_avatar_id(list_function: Callable, show_max_avatars: int) -> tuple[str, str]:
    """Select which avatar to download from available asset ids"""
    avatars = get_ready_assets(list_function=list_function, show_max_assets=show_max_avatars)
    # No avatars available for this account
    if len(avatars) == 0:
        raise click.ClickException("You have no created avatars to download.")
    # Print avatars available for selection
    prompt_text = "\n\n{:<8}{:<30}{:<40}{:<20}{:<20}\n".format("Number", "Name", "Asset ID", "Created from", "Timestamp")
    for i, avatar in enumerate(avatars):
        prompt_text += "-"*120 + "\n"
        prompt_text += "{:<8}{:<30}{:<40}{:<20}{:<20}\n".format(i, avatar["name"], avatar["assetID"], avatar["origin"], avatar["timestamp"])
    log.info(prompt_text)
    avatar_number = click.prompt('Number of avatar to download', type=int)
    return avatars[avatar_number]["assetID"], avatars[avatar_number]["name"]

def select_motion_id(list_function: Callable, show_max_motions: int, motion_number_string: str = "") -> tuple[str, str]:
    """Select which motion to select for blending from available asset ids"""
    motions = get_ready_assets(list_function=list_function, show_max_assets=show_max_motions)
    # No motions available for this account
    if len(motions) == 0:
        raise click.ClickException("You have not created any motions to select.")
    # Print avatars available for selection
    prompt_text = "\n\n{:<8}{:<40}{:<20}{:<20}\n".format("Number", "Asset ID", "Created from", "Timestamp")
    for i, motion in enumerate(motions):
        prompt_text += "-"*120 + "\n"
        prompt_text += "{:<8}{:<40}{:<20}{:<20}\n".format(i, motion["assetID"], motion["origin"], motion["timestamp"])
    log.info(prompt_text)
    motion_number = click.prompt(f'Select {motion_number_string}motion', type=int)
    return motions[motion_number]["assetID"]

class TimeoutTracker(object):
    """Helper class for tracking timeout length"""
    def __init__(self, timeout_length):
        self.start_time = time.time()
        self.timeout = timeout_length
    def reset(self):
        self.start_time = time.time()
    def timed_out(self):
        return time.time() - self.start_time > self.timeout*60
    def is_active(self):
        return not self.timed_out()

def get_timestamp():
    """Get timestamp"""
    return datetime.now().strftime("%Y%m%d%H%M%S")

class Uploader:
    """Class for uploading files that can be mocked"""
    def upload(self, file_to_upload: str, upload_url: str) -> None:
        """Upload an image to given url"""
        content_type = MIME_TYPES[file_to_upload.split(".")[-1].lower()]
        with open(file_to_upload, 'rb') as input:
            upload_response = requests.put(upload_url, data=input, headers={'Content-Type': content_type})
        upload_response.raise_for_status()


class LocalUploader:
    """Class for uploading files that can be mocked"""

    def upload(self, file_to_upload: str, upload_url: str) -> None:
        # TODO(dheid)
        # Due to time constraints and to enable debugging this is here for now
        # Make this configurable and move it back into the Uploader above
        # Should not always be the case, and only be done if we target the local mcme stack
        # See mcme/batch_create.py line 71 and replace that with Uploader() when done
        upload_url = upload_url.replace("localstack", "localhost")
        """Upload an image to given url"""
        content_type = MIME_TYPES[file_to_upload.split(".")[-1].lower()]
        with open(file_to_upload, 'rb') as input:
            upload_response = requests.put(upload_url, data=input, headers={'Content-Type': content_type})
        upload_response.raise_for_status()


def get_measurements_dict(
    height,
    weight,
    bust_girth,
    ankle_girth,
    thigh_girth,
    waist_girth,
    armscye_girth,
    top_hip_girth,
    neck_base_girth,
    shoulder_length,
    lower_arm_length,
    upper_arm_length,
    inside_leg_height
    ) -> dict[str, float]:
    """Helper for assembling input measurements"""
    measurements = {
                        "Height": height,
                        "Weight": weight,
                        "Bust_girth": bust_girth,
                        "Ankle_girth": ankle_girth,
                        "Thigh_girth": thigh_girth,
                        "Waist_girth": waist_girth,
                        "Armscye_girth": armscye_girth,
                        "Top_hip_girth": top_hip_girth,
                        "Neck_base_girth": neck_base_girth,
                        "Shoulder_length": shoulder_length,
                        "Lower_arm_length": lower_arm_length,
                        "Upper_arm_length": upper_arm_length,
                        "Inside_leg_height": inside_leg_height
                    }
    for key, value in measurements.copy().items():
        if value is None:
            del measurements[key]
    return measurements




import time
from typing import Optional, Dict, Union
import click
import openapi_client as client
from openapi_client import PostgresqlAssetState as AssetState
from openapi_client import PostgresqlBuildState as BuildState
from openapi_client import SchemasMotionBlendMotion
from .logger import log
from .helpers import TimeoutTracker, Uploader


def from_betas(
        gender: Optional[client.EnumsGender], 
        betas: list[float], 
        name: str, 
        model_version: Optional[client.EnumsModelVersion],
        api_instance: client.CreateAvatarsFromBetasApi
        ) -> str:
    """Create avatar from betas."""
    betas_request = client.SchemasBetasAvatarRequest(
        betas=betas,
        gender=gender,
        name=name,
        modelVersion=model_version)

    try:
        # Creates avatar from betas
        api_response = api_instance.create_avatar_from_betas(betas_request)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarsFromBetasApi->create_avatar_from_betas: %s\n" % e) from e
    if api_response.data is None or api_response.data.attributes is None:
        raise click.ClickException("Creating avatar from betas response came back empty")
    log.info(f"Creating an avatar from betas finished with state {AssetState(api_response.data.attributes.state).name}")
    return str(api_response.data.id)


def get_processing_state(api_instance: client.AvatarsApi, asset_id: str) -> AssetState:
    """List avatar to retrieve its state"""
    try:
        # List one avatar
        api_response = api_instance.describe_avatar(asset_id)
    except Exception as e:
        raise click.ClickException("Exception when calling AvatarsApi->describe_avatar: %s\n" % e) from e
    if api_response.data is None or \
        api_response.data.attributes is None:
        raise click.ClickException("Response came back empty")
    state = api_response.data.attributes.state
    if state == AssetState.ERROR:
        raise click.ClickException("Processing finished with state ERROR")
    else:
        return AssetState(state)


def request_avatar_from_images(api_instance: client.CreateAvatarFromImagesApi) -> str:
    """Initiate avatar from images creation"""
    try:
        api_response = api_instance.create_avatar_from_images()
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromImagesApi->create_avatar_from_images: %s\n" % e) from e
    if api_response.data is None:
        raise click.ClickException("Initiating avatar from images response came back empty")
    asset_id = str(api_response.data.id)
    log.info(f"AssetID: {asset_id}")
    return asset_id


def request_avatar_from_video(api_instance: client.CreateAvatarFromVideoApi) -> str:
    """Initiate avatar from video creation"""
    try:
        api_response = api_instance.create_avatar_from_video()
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromVideoApi->create_avatar_from_video: %s\n" % e) from e
    if api_response.data is None:
        raise click.ClickException("Initiating avatar from video response came back empty")
    asset_id = str(api_response.data.id)
    log.info(f"AssetID: {asset_id}")
    return asset_id

def request_avatar_from_scans(api_instance: client.CreateAvatarFromScansApi) -> str:
    """Initiate avatar from scans creation"""
    try:
        api_response = api_instance.create_avatar_from_scans()
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromScansApi->create_avatar_from_scans: %s\n" % e) from e
    if api_response.data is None:
        raise click.ClickException("Initiating avatar from scans response came back empty")
    asset_id = str(api_response.data.id)
    log.info(f"AssetID: {asset_id}")
    return asset_id

def request_image_upload(
        api_instance: client.CreateAvatarFromImagesApi,
        asset_id: str
        ) -> str:
    """Request image upload URL for avatar creation"""
    try:
        api_response = api_instance.upload_image_to_avatar(asset_id)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromImagesApi->upload_image_to_avatar: %s\n" % e) from e

    try:
        return str(api_response.data.attributes.url.path)
    except AttributeError as e:
        raise click.ClickException("Requesting image upload response came back empty") from e

def request_video_upload(
        api_instance: client.CreateAvatarFromVideoApi,
        asset_id: str
        ) -> str:
    """Request video upload URL for avatar creation"""
    try:
        api_response = api_instance.upload_video_to_avatar(asset_id)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromVideoApi->upload_video_to_avatar: %s\n" % e) from e

    try:
        return str(api_response.data.attributes.url.path)
    except AttributeError as e:
        raise click.ClickException("Requesting video upload response came back empty") from e

def request_scan_upload(
        api_instance: client.CreateAvatarFromScansApi,
        asset_id: str
        ) -> str:
    """Request image upload URL for avatar creation"""
    try:
        api_response = api_instance.upload_mesh_to_avatar(asset_id)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromScansApi->upload_mesh_to_avatar: %s\n" % e) from e
    if api_response.data is None or \
            api_response.data.attributes is None or \
            api_response.data.attributes.url is None:
        raise click.ClickException("Requesting image upload response came back empty")
    return str(api_response.data.attributes.url.path)


def from_images(gender: Optional[client.EnumsGender],
                name: str,
                input: str,
                height: int,
                weight: int,
                image_mode: str,
                api_instance_images: client.CreateAvatarFromImagesApi,
                api_instance_avatars: client.AvatarsApi,
                uploader: Uploader,
                timeout: int
                ) -> str:
    """Create avatar from images."""

    asset_id = request_avatar_from_images(api_instance=api_instance_images)

    upload_url = request_image_upload(api_instance=api_instance_images, asset_id=asset_id)

    uploader.upload(file_to_upload=input, upload_url=upload_url)

    # Fit to images
    afi_inputs = client.DocschemasDocAFIInputs(avatarname=name, 
                                                       gender=gender, 
                                                       height=height, 
                                                       weight=weight, 
                                                       imageMode=image_mode)
    try:
        api_response = api_instance_images.avatar_fit_to_images(asset_id, afi_inputs)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromImagesApi->avatar_fit_to_images: %s\n" % e) from e
    if api_response.data is None or api_response.data.attributes is None:
            raise click.ClickException("Fitting to images response came back empty")
    
    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    afi_state = ""
    while afi_state != AssetState.READY and not timeout_tracker.timed_out():
        afi_state = get_processing_state(api_instance_avatars, asset_id)
        log.info(f"Avatar from images processing state: {afi_state.name}" if afi_state != AssetState.READY \
                 else f"Avatar from images processing finished with state {afi_state.name}")
        time.sleep(5)
    if afi_state != AssetState.READY:
        # image creation didn't finish before it timed out
        raise click.ClickException("Avatar from images creation timed out.")
    return asset_id

def from_video(gender: Optional[client.EnumsGender],
                name: str,
                input: str,
                api_instance_images: client.CreateAvatarFromVideoApi,
                api_instance_avatars: client.AvatarsApi,
                uploader: Uploader,
                timeout: int
                ) -> str:
    """
    Create avatar from a single video.
    Uses MCME openai client to create an avatar from a video file.
    First creates an empty avatar asset and then uploads the video file to the asset.
    Then calls the fit to video endpoint to start the process.

    :param client.EnumsGender gender: Gender of the created avatar
    :param str name: Name of the created avatar
    :param str input: Path to the video file
    :param client.CreateAvatarFromVideoApi api_instance_images: API instance for creating avatar from video
    :param client.AvatarsApi api_instance_avatars: API instance for avatars endpoints
    :param Uploader uploader: File Uploader instance
    :param int timeout: Timeout in seconds
    :return: Asset ID of the created avatar
    :rtype: str
    :raises click.ClickException: If any of the API calls fail or the processing times timed_out
    """

    asset_id = request_avatar_from_video(api_instance=api_instance_images)

    upload_url = request_video_upload(api_instance=api_instance_images, asset_id=asset_id)

    uploader.upload(file_to_upload=input, upload_url=upload_url)

    # Fit to video
    afv_inputs = client.DocschemasDocAFIInputs(avatarname=name, 
                                                       gender=gender)
    try:
        api_response = api_instance_images.avatar_fit_to_video(asset_id, afv_inputs)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromImagesApi->avatar_fit_to_images: %s\n" % e) from e

    if api_response.data is None or api_response.data.attributes is None:
            raise click.ClickException("Fitting to video response came back empty")
    
    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    afv_state = AssetState.AWAITING_PROCESSING

    # TODO(dheid) - refactor this out into a helper function
    #def ready_within_timeout(api_instance, asset_id, timeout):
    #make tracker
    #while active:
    #   if state == ready: 
    #     return True
    #    sleep
    #else:
    #    return False
    # This needs a bit of extra refactoring, since as of right now it only works for avatar AssetState
    # Each asset type has its own kind of api instance with its own method name to fetch the asset properties
    while timeout_tracker.is_active():
        if get_processing_state(api_instance_avatars, asset_id) == AssetState.READY:
            log.info(f"Avatar from video processing finished with state {AssetState.READY.name}")
            break
        log.info(f"Avatar from video processing state: {afv_state.name}")
        time.sleep(5)
    else:
        raise click.ClickException("Avatar from video creation timed out.")

    return asset_id

def from_scans(gender: Optional[client.EnumsGender],
                name: str,
                input: str,
                init_pose: str,
                up_axis: str,
                look_axis: str,
                input_units: str,
                api_instance_scans: client.CreateAvatarFromScansApi,
                api_instance_avatars: client.AvatarsApi,
                uploader: Uploader,
                timeout: int
                ) -> str:
    """Create avatar from images."""

    asset_id = request_avatar_from_scans(api_instance=api_instance_scans)

    upload_url = request_scan_upload(api_instance=api_instance_scans, asset_id=asset_id)

    uploader.upload(file_to_upload=input, upload_url=upload_url)

    # Fit to images
    afs_inputs = client.DocschemasDocAFSInputs(avatarname=name,
                                                       gender=gender,
                                                       initPose=init_pose,
                                                       upAxis=up_axis,
                                                       lookAxis=look_axis,
                                                       inputUnits=input_units
                                                       )
    try:
        api_response = api_instance_scans.avatar_fit_to_scans(asset_id, afs_inputs)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromScansApi->avatar_fit_to_scans: %s\n" % e) from e

    if api_response.data is None or api_response.data.attributes is None:
        raise click.ClickException("Fitting to scans response came back empty")
    
    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    afs_state = AssetState.AWAITING_PROCESSING
    while afs_state != AssetState.READY and not timeout_tracker.timed_out():
        afs_state = get_processing_state(api_instance_avatars, asset_id)
        log.info(f"Avatar from scans processing state: {afs_state.name}" if afs_state != AssetState.READY \
                 else f"Avatar from scans processing finished with state {afs_state.name}")
        time.sleep(5)
    if afs_state != AssetState.READY:
        # image creation didn't finish before it timed out
        raise click.ClickException("Avatar from scans creation timed out.")
    return asset_id


def from_measurements(gender: Optional[client.EnumsGender],
                name: str,
                measurements: Optional[Dict[str, Union[float, int]]],
                model_version: Optional[client.EnumsModelVersion],
                api_instance_from_measurements: client.CreateAvatarFromMeasurementsApi,
                api_instance_avatars: client.AvatarsApi,
                timeout: int
                ) -> str:
    """Create avatar from measurements."""
    afm_inputs = client.SchemasMeasurementAvatarRequest(gender=gender,
                                                                name=name,
                                                                measurements=measurements,
                                                                modelVersion=model_version)

    try:
        api_response = api_instance_from_measurements.avatar_from_measurements(afm_inputs)
    except Exception as e:
        raise click.ClickException("Exception when calling CreateAvatarFromMeasurementsApi->avatar_from_measurements: %s\n" % e) from e
    if api_response.data is None:
            raise click.ClickException("Avatar from measurements response came back empty")
    asset_id = str(api_response.data.id)
    
    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    afm_state = AssetState.AWAITING_PROCESSING
    while afm_state != AssetState.READY and not timeout_tracker.timed_out():
        afm_state = get_processing_state(api_instance_avatars, asset_id)
        if afm_state == AssetState.READY:
            log.info(f"Avatar from measurements processing finished with state {afm_state.name}")
        else:
            log.info(f"Avatar from measurements processing state: {afm_state.name}")
            time.sleep(5)
    if afm_state != AssetState.READY:
        # afm creation didn't finish before it timed out
        raise click.ClickException("Avatar from measurements creation timed out.")
    return asset_id

def from_smpl(
        smpl_file: str,
        api_instance: client.AvatarsApi,
        name: str,
        uploader: Uploader,
        timeout: int
        ) -> str:
    """"Create avatar from .smpl file"""
    smpl_inputs = client.DocschemasDocCreateFromSMPLRequest(name=name)
    try:
        api_response = api_instance.create_from_smpl(smpl_request=smpl_inputs)
    except Exception as e:
        raise click.ClickException("Exception when calling AvatarsApi->create_from_smpl: %s\n" % e) from e
    if api_response.data is None or \
        api_response.data.attributes is None or \
        api_response.data.attributes.url is None:
        raise click.ClickException("Initiating avatar from smpl response came back empty")
    # Get asset id of target avatar asset
    asset_id = str(api_response.data.id)
    # Upload .smpl file
    upload_url = str(api_response.data.attributes.url.path)
    uploader.upload(file_to_upload=smpl_file, upload_url=upload_url)

    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    state = AssetState.AWAITING_PROCESSING
    while state != AssetState.READY and not timeout_tracker.timed_out():
        state = get_processing_state(api_instance, asset_id)
        if state == AssetState.READY:
            log.info(f"Avatar from text processing finished with state {state.name}")
        else:
            log.info(f"Avatar from text processing state: {state.name}")
            time.sleep(5)
    if state != AssetState.READY:
        # avatar from smpl creation didn't finish before it timed out
        raise click.ClickException("Avatar from text creation timed out.")
    return asset_id


def blend_motions(
        source_avatar_id: str,
        motion_id_1: str,
        motion_id_2: str,
        shape_parameters: Optional[list[float]], 
        gender: Optional[client.EnumsGender], 
        name: str,
        api_instance: client.BlendMotionsApi,
        avatar_api_instance: client.AvatarsApi,
        timeout: int
        ) -> str:
    """Create avatar with blended motions attached."""
    blend_motions_request = client.SchemasMotionBlendRequest(
        avatarName=name, 
        gender=gender, 
        motions=[SchemasMotionBlendMotion.from_dict({"motionId": motion_id_1}), SchemasMotionBlendMotion.from_dict({"motionId": motion_id_2})], 
        shapeParameters=shape_parameters, 
        sourceAvatarID=source_avatar_id)

    try:
        # Creates avatar with blended motions attached
        api_response = api_instance.blend_motions(blend_motions_request)
    except Exception as e:
        raise click.ClickException("Exception when calling BlendMotionsApi->blend_motions: %s\n" % e) from e
    if api_response.data is None or api_response.data.attributes is None:
        raise click.ClickException("Blend motion response came back empty")
    asset_id = str(api_response.data.id)
    
    # Wait for processing to finish
    timeout_tracker = TimeoutTracker(timeout)
    blend_motions_state = AssetState.AWAITING_PROCESSING
    while blend_motions_state != AssetState.READY and not timeout_tracker.timed_out():
        blend_motions_state = get_processing_state(avatar_api_instance, asset_id)
        if blend_motions_state == AssetState.READY:
            log.info(f"Creating an avatar with blended motion finished with state {blend_motions_state.name}")
        else:
            log.info(f"Blend motions processing state: {blend_motions_state.name}")
            time.sleep(5)
    if blend_motions_state != AssetState.READY:
        # afm creation didn't finish before it timed out
        raise click.ClickException("Blend motions timed out.")
    return asset_id


def try_get_download_url(
        api_instance: client.AvatarsApi,
        asset_id: str,
        input: client.DocschemasDocExportInputs) -> Optional[str]:
    """Initiate avatar export and call export endpoint to check on state"""
    try:
        # Call export avatar endpoint
        api_response = api_instance.export_avatar(asset_id, input)
    except Exception as e:
        raise click.ClickException("Exception when calling AvatarsApi->export_avatar: %s\n" % e) from e
    if api_response.data is None or \
        api_response.data.attributes is None or \
        api_response.data.attributes.url is None:
        raise click.ClickException("Export avatar response came back empty")
    # get current processing state and download url from api response
    state = api_response.data.attributes.state
    download_url = str(api_response.data.attributes.url.path)
    if state == AssetState.READY:
            # export finished with state ready, return results
            log.info(f"Exporting the created avatar finished with state {AssetState(state).name}")
            return download_url
    elif state == AssetState.ERROR:
        raise click.ClickException("Exporting finished with state ERROR")
    else:
        log.info(f"Exporting avatar state: {AssetState(state).name}")
        return None
        
        
def export( 
        asset_id: str, 
        download_format: str, 
        pose: str, 
        animation: str, 
        compatibility_mode: str,
        api_instance: client.AvatarsApi,
        timeout: int
        ) -> str:
    """Export avatar"""
    input = client.DocschemasDocExportInputs(
        format=download_format,
        pose=pose,
        anim=animation,
        compatibilityMode=compatibility_mode)
    timeout_tracker = TimeoutTracker(timeout)
    download_url =  None
    while not download_url and not timeout_tracker.timed_out():
        download_url = try_get_download_url(api_instance, asset_id, input)
        # if state is still processing or awaiting processing, call export enpoint again after waiting a bit
        time.sleep(5)
    if download_url is None:
        # export didn't return anything before it timed out
        raise click.ClickException("Export timed out.")
    return download_url



# flake8: noqa

# import apis into api package
from openapi_client.api.assets_api import AssetsApi
from openapi_client.api.avatars_api import AvatarsApi
from openapi_client.api.blend_motions_api import BlendMotionsApi
from openapi_client.api.create_avatar_from_images_api import CreateAvatarFromImagesApi
from openapi_client.api.create_avatar_from_scans_api import CreateAvatarFromScansApi
from openapi_client.api.create_avatar_from_video_api import CreateAvatarFromVideoApi
from openapi_client.api.create_avatar_from_measurements_api import CreateAvatarFromMeasurementsApi
from openapi_client.api.create_avatars_from_betas_api import CreateAvatarsFromBetasApi
from openapi_client.api.create_scene_from_video_api import CreateSceneFromVideoApi
from openapi_client.api.credit_bundles_api import CreditBundlesApi
from openapi_client.api.helpers_api import HelpersApi
from openapi_client.api.images_api import ImagesApi
from openapi_client.api.info_api import InfoApi
from openapi_client.api.measure_avatar_api import MeasureAvatarApi
from openapi_client.api.mesh_api import MeshApi
from openapi_client.api.motions_api import MotionsApi
from openapi_client.api.search_motions_api import SearchMotionsApi
from openapi_client.api.subscription_api import SubscriptionApi
from openapi_client.api.user_api import UserApi
from openapi_client.api.videos_api import VideosApi




# coding: utf-8

"""
    Meshcapade Me API

    Welcome to the age of Avatars, introducing the Meshcapade Me API. The no-friction avatar-creation platform for all your avatar needs. Create accurate digital doubles from any source of data in a unified 3D body format for every industry. Built on our <a href=\"https://meshcapade.com/SMPL\" target=\"_blank\">patented avatar</a> technology, our API allows you to create and edit avatars from images, videos, measurements, scans, and more. # Introduction The Meshcapade Me API is a RESTful API that allows you to create and edit avatars from images, measurements, scans, and more. All API replies adhere to the  <a href=\"https://jsonapi.org/format/\" target=\"_blank\">JSON:API</a> schema guidelines. Currently the API is in beta and is subject to change. Thus, not all ways to create avatars are available yet. We are working hard to add more ways to create avatars and will update this documentation accordingly. The API allows you to create avatars </br> - From <a href=\"#post-/avatars/create/from-images\" target=\"_blank\">images</a> </br> - From <a href=\"#post-/avatars/create/from-video\" target=\"_blank\">video</a> </br> - From <a href=\"#post-/avatars/create/from-scans\" target=\"_blank\">3D scans</a> </br> - From <a href=\"#post-/avatars/create/from-measurements\" target=\"_blank\">measurements</a> </br> # Quickstart To get started, sign up for a free account at <a href=\"https://me.meshcapade.com\" target=\"_blank\">me.meshcapade.com</a></br> We recommend using our <a href=\"https://www.postman.com/downloads/\" target=\"_blank\">Postman</a> collection to conveniently explore the API </br> <div style=\"margin-top: 16px;\"><a href=\"https://www.postman.com/cloudy-meadow-883625/workspace/meshcapade/overview\"><img src=\"https://run.pstmn.io/button.svg\" alt=\"Run in Postman\"></a></div></br>  # How-To <a href=\"https://medium.com/meshcapade/streamline-avatar-creation-with-meshcapade-me-api-from-one-image-to-an-accurate-avatar-in-seconds-b8ca4f15b9a8\" target=\"_blank\">Create an avatar from a single image (Medium)</a> </br> <a href=\"https://medium.com/meshcapade/measurements-meet-imagination-creating-accurate-3d-avatars-with-meshcapades-api-9a6ec5029793\" target=\"_blank\">Create an avatar from measurements (Medium)</a> # API Categories The API is organized into the following main categories: - Assets </br> Endpoint for listing assets of multiple types. - Avatars </br> Endpoints for listing, downloading, and deleting avatars. - Mesh </br> Endpoints for listing, downloading, and deleting exported meshes. - Images </br> Endpoints for listing, uploading, and deleting images related to avatars. - Create Avatar from images </br> Endpoints to initiate and complete avatar creation from images. - Create Avatar from measurements </br> Endpoint to initiate avatar creation from body measurements. - Create Avatar from scans </br> Endpoints to initiate and complete avatar creation from 3d body scans. - Create Avatar from betas </br> Endpoint to create an avatar from SMPL based beta shape parameters. # Error codes When something goes wrong, the API replies with an additional error code  - `asset_not_found` The requested asset either does not exist, or is not owned by the user (404)  </br> - `too_many_images` The image limit that can be uploaded for avatars from images as been exceeded. (400) </br> - `already_started` A process that already has been requested cannot be started again. (400) </br> - `no_images`  [POST /avatars/create/from-images](#post-/avatars/create/from-images) can only be started with at least one image uploaded  (400) </br> - `inputs_not_ready` Running `/avatars/create/xxxx` endpoint require the inputs to be uploaded (400) </br> - `uuid_invalid_format` Asset ID is in a non-UUID format (400) </br> - ` missing_parameters` Not all required parameters have been supplied for the request (400) </br> - `unauthorized` Trying to access an asset the user does not own, or endpoints that the user is not authorized to call (400) </br> - `too_many_builds` User processing rate has been exceeded. Only one computation heavy process can be started at a time (429) </br> - `asset_not_ready` Cannot call request on an asset which is not in a READY state (400) </br>  # Integration When integrating with our API, you may encounter Cross-Origin Resource Sharing (CORS) issues during deployment. </br> For security reasons, our API does not support direct communication from your frontend. </br> Instead, we recommend that you connect to api.meshcapade through your own backend server. </br> This approach not only mitigates CORS-related challenges but also enhances the overall security of your application. </br> By handling API requests server-side, you can ensure smoother and safer integration. </br> If you encounter any issues during integration, please reach out to us at support@meshcapade.com </br>  # noqa: E501

    The version of the OpenAPI document: v1.20
    Contact: support@meshcapade.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


import re  # noqa: F401
import io
import warnings

from pydantic import validate_arguments, ValidationError
from typing_extensions import Annotated

from typing_extensions import Annotated
from pydantic import Field, StrictStr, conint

from typing import Optional

from openapi_client.models.docschemas_doc_base_jsonapi_response import DocschemasDocBaseJSONAPIResponse

from openapi_client.api_client import ApiClient
from openapi_client.api_response import ApiResponse
from openapi_client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class MotionsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def delete_motion(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of avatar to delete")], **kwargs) -> None:  # noqa: E501
        """Delete Motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_motion(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of avatar to delete (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the delete_motion_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.delete_motion_with_http_info(asset_id, **kwargs)  # noqa: E501

    @validate_arguments
    def delete_motion_with_http_info(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of avatar to delete")], **kwargs) -> ApiResponse:  # noqa: E501
        """Delete Motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_motion_with_http_info(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of avatar to delete (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """

        _params = locals()

        _all_params = [
            'asset_id'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_motion" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params['asset_id']:
            _path_params['assetID'] = _params['asset_id']


        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # authentication setting
        _auth_settings = ['Bearer']  # noqa: E501

        _response_types_map = {}

        return self.api_client.call_api(
            '/motions/{assetID}', 'DELETE',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def describe_motions(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of avatar to list")], **kwargs) -> None:  # noqa: E501
        """List one motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.describe_motions(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of avatar to list (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the describe_motions_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.describe_motions_with_http_info(asset_id, **kwargs)  # noqa: E501

    @validate_arguments
    def describe_motions_with_http_info(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of avatar to list")], **kwargs) -> ApiResponse:  # noqa: E501
        """List one motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.describe_motions_with_http_info(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of avatar to list (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """

        _params = locals()

        _all_params = [
            'asset_id'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method describe_motions" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params['asset_id']:
            _path_params['assetID'] = _params['asset_id']


        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # authentication setting
        _auth_settings = ['Bearer']  # noqa: E501

        _response_types_map = {}

        return self.api_client.call_api(
            '/motions/{assetID}', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def list_motions(self, limit : Annotated[Optional[conint(strict=True, ge=1)], Field(description="Paginated motion limit amount (Default 20)")] = None, page : Annotated[Optional[conint(strict=True, ge=1)], Field(description="Pagination")] = None, **kwargs) -> DocschemasDocBaseJSONAPIResponse:  # noqa: E501
        """List all motions  # noqa: E501

        Lists all motions the user owns.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_motions(limit, page, async_req=True)
        >>> result = thread.get()

        :param limit: Paginated motion limit amount (Default 20)
        :type limit: int
        :param page: Pagination
        :type page: int
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: DocschemasDocBaseJSONAPIResponse
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the list_motions_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.list_motions_with_http_info(limit, page, **kwargs)  # noqa: E501

    @validate_arguments
    def list_motions_with_http_info(self, limit : Annotated[Optional[conint(strict=True, ge=1)], Field(description="Paginated motion limit amount (Default 20)")] = None, page : Annotated[Optional[conint(strict=True, ge=1)], Field(description="Pagination")] = None, **kwargs) -> ApiResponse:  # noqa: E501
        """List all motions  # noqa: E501

        Lists all motions the user owns.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_motions_with_http_info(limit, page, async_req=True)
        >>> result = thread.get()

        :param limit: Paginated motion limit amount (Default 20)
        :type limit: int
        :param page: Pagination
        :type page: int
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(DocschemasDocBaseJSONAPIResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'limit',
            'page'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_motions" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get('limit') is not None:  # noqa: E501
            _query_params.append(('limit', _params['limit']))

        if _params.get('page') is not None:  # noqa: E501
            _query_params.append(('page', _params['page']))

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # authentication setting
        _auth_settings = ['OAuth2Password', 'Bearer']  # noqa: E501

        _response_types_map = {
            '200': "DocschemasDocBaseJSONAPIResponse",
            '400': "DocschemasDocErrorResponse",
            '401': "DocschemasDocErrorResponse",
            '404': "DocschemasDocErrorResponse",
        }

        return self.api_client.call_api(
            '/motions', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def upload_motion(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of motion to upload")], **kwargs) -> None:  # noqa: E501
        """Generate upload link for a motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.upload_motion(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of motion to upload (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """
        kwargs['_return_http_data_only'] = True
        if '_preload_content' in kwargs:
            raise ValueError("Error! Please call the upload_motion_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data")
        return self.upload_motion_with_http_info(asset_id, **kwargs)  # noqa: E501

    @validate_arguments
    def upload_motion_with_http_info(self, asset_id : Annotated[StrictStr, Field(..., description="assetID of motion to upload")], **kwargs) -> ApiResponse:  # noqa: E501
        """Generate upload link for a motion  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.upload_motion_with_http_info(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: assetID of motion to upload (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the 
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """

        _params = locals()

        _all_params = [
            'asset_id'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method upload_motion" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params['asset_id']:
            _path_params['assetID'] = _params['asset_id']


        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # authentication setting
        _auth_settings = ['Bearer']  # noqa: E501

        _response_types_map = {}

        return self.api_client.call_api(
            '/motions/{assetID}/upload', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))



import click
from .logger import log
from dataclasses import dataclass

@dataclass
class User:
    email: str
    credits: int

def request_user_info(user_api_instance) -> User:
    """
    Request user info from the user.
    """
    try:
        api_response_user = user_api_instance.describe_user()
        if api_response_user.data is None or api_response_user.data.attributes is None:
            raise click.ClickException("User info response came back empty")
        return User(email=api_response_user.data.attributes.email, credits=api_response_user.data.attributes.credits)
    except Exception as e:
        log.info("Exception when calling UserApi->describe_user: %s\n" % e)

    return User(email="", credits=0)




import os
import tempfile

import click
from .helpers import Uploader
from .avatars import from_smpl
import requests
import openapi_client as client
from .logger import log
from openapi_client import PostgresqlBuildState as BuildState
from smplcodec import SMPLCodec
import dataclasses

motion_parameters = ["body_translation",
                    "body_pose",
                    "head_pose",
                    "left_hand_pose",
                    "right_hand_pose"]

class TMRMotion():
    """
    Temporary motion retrieved by text prompt using the /motions/search API endpoint.
    The motion can be found by text prompt and trimmed to the relevant frames, creating a temporary .smpl file.
    """

    def __init__(self,  prompt: str) -> None:
        """Initializing motion"""
        self.prompt = prompt


    def find_motion(self, api_instance: client.SearchMotionsApi) -> None:
        """Finding motion and downloading motion .smpl file as a temporary file"""
        motion_options = client.ServicesSearchMotionsOptions(
        num_motions=1,
        text=self.prompt)
        try:
            # Search for motion using prompt
            api_response = api_instance.submit_search_motions(motion_options)
        except Exception as e:
            raise click.ClickException("Exception when calling SearchMotionsApi->submit_search_motions: %s\n" % e) from e
        if api_response.data is None or api_response.data.attributes is None:
            raise click.ClickException("Searching for motion response came back empty")
        log.info(f"Creating motion from text finished with state {BuildState(api_response.data.attributes.state).name}")
        if api_response.included is None or \
            len(api_response.included) == 0 or \
            api_response.included[0] is None or \
            api_response.included[0].attributes is None or \
            api_response.data is None or \
            api_response.data.attributes is None or \
            api_response.data.attributes.result is None or \
            api_response.data.attributes.result.motions is None or \
            len(api_response.data.attributes.result.motions) == 0 or \
            api_response.data.attributes.result.motions[0] is None or \
            api_response.data.attributes.result.motions[0].start_time is None or \
            api_response.data.attributes.result.motions[0].end_time is None:
                raise click.ClickException("No motion found.")
        
        # Temporarily downloading motion .smpl file
        motion_download_url = api_response.included[0].attributes.url.path
        log.debug(f"Motion found: {api_response.data.attributes.result.motions[0].path_key}")
        with tempfile.NamedTemporaryFile(delete=False, suffix='.smpl') as file:
            self.file = file.name
            try:
                stream = requests.get(motion_download_url, stream=True, timeout=60)
                stream.raise_for_status()
            except requests.exceptions.HTTPError as err:
                raise click.ClickException(str(err)) from err
            for chunk in stream.iter_content(chunk_size=1024 * 1024):
                file.write(chunk)

        # Save start and end times of motion for later trimming
        self.start = api_response.data.attributes.result.motions[0].start_time
        self.end = api_response.data.attributes.result.motions[0].end_time


    def trim(self, start=None, end=None) -> str:
        """Trimming motion to relevant part as indicated in search motions response"""
        if start == None:
            start = self.start
        if end == None:
            end = self.end
        motion = SMPLCodec.from_file(self.file)

        # set original file as trimmed file if trimming is unnecessary
        if end == motion.frame_count/motion.frame_rate:
            log.debug("Motion trimming not necessary.")
            self.trimmed_file = self.file
            return self.trimmed_file
        
        # else, write trimmed temp file
        first_frame = int(start * motion.frame_rate)
        last_frame = int(end * motion.frame_rate)
        # update frame count to prevent smplcodec validation from failing
        setattr(motion, "frame_count", last_frame-first_frame)
        for attr in motion_parameters:
            if hasattr(motion, attr):
                frames = getattr(motion, attr)
                trimmed_frames = frames[first_frame:last_frame]
                setattr(motion,attr,trimmed_frames)
                
        
        temp = tempfile.NamedTemporaryFile(delete=False, suffix='_trimmed.smpl')
        try:
            self.trimmed_file = temp.name
            # remove codec_version since it is not compatible with the API (yet)
            del motion.__dataclass_fields__["codec_version"]
            motion.write(temp.name)
        finally:
            temp.close()
        return self.trimmed_file


    def cleanup(self):
        """"Deleting temporary .smpl file containing the downloaded motion"""
        os.remove(self.file)
        if os.path.isfile(self.trimmed_file):
            os.remove(self.trimmed_file)



from dataclasses import dataclass
from typing import Optional


@dataclass
class ExportParameters:
    """Class for defining export parameters."""
    download_format: str
    pose: str
    animation: str
    compatibility_mode: str
    out_file: str



# coding: utf-8

"""
    Meshcapade Me API

    Welcome to the age of Avatars, introducing the Meshcapade Me API. The no-friction avatar-creation platform for all your avatar needs. Create accurate digital doubles from any source of data in a unified 3D body format for every industry. Built on our <a href=\"https://meshcapade.com/SMPL\" target=\"_blank\">patented avatar</a> technology, our API allows you to create and edit avatars from images, videos, measurements, scans, and more. # Introduction The Meshcapade Me API is a RESTful API that allows you to create and edit avatars from images, measurements, scans, and more. All API replies adhere to the  <a href=\"https://jsonapi.org/format/\" target=\"_blank\">JSON:API</a> schema guidelines. Currently the API is in beta and is subject to change. Thus, not all ways to create avatars are available yet. We are working hard to add more ways to create avatars and will update this documentation accordingly. The API allows you to create avatars </br> - From <a href=\"#post-/avatars/create/from-images\" target=\"_blank\">images</a> </br> - From <a href=\"#post-/avatars/create/from-video\" target=\"_blank\">video</a> </br> - From <a href=\"#post-/avatars/create/from-scans\" target=\"_blank\">3D scans</a> </br> - From <a href=\"#post-/avatars/create/from-measurements\" target=\"_blank\">measurements</a> </br> # Quickstart To get started, sign up for a free account at <a href=\"https://me.meshcapade.com\" target=\"_blank\">me.meshcapade.com</a></br> We recommend using our <a href=\"https://www.postman.com/downloads/\" target=\"_blank\">Postman</a> collection to conveniently explore the API </br> <div style=\"margin-top: 16px;\"><a href=\"https://www.postman.com/cloudy-meadow-883625/workspace/meshcapade/overview\"><img src=\"https://run.pstmn.io/button.svg\" alt=\"Run in Postman\"></a></div></br>  # How-To <a href=\"https://medium.com/meshcapade/streamline-avatar-creation-with-meshcapade-me-api-from-one-image-to-an-accurate-avatar-in-seconds-b8ca4f15b9a8\" target=\"_blank\">Create an avatar from a single image (Medium)</a> </br> <a href=\"https://medium.com/meshcapade/measurements-meet-imagination-creating-accurate-3d-avatars-with-meshcapades-api-9a6ec5029793\" target=\"_blank\">Create an avatar from measurements (Medium)</a> # API Categories The API is organized into the following main categories: - Assets </br> Endpoint for listing assets of multiple types. - Avatars </br> Endpoints for listing, downloading, and deleting avatars. - Mesh </br> Endpoints for listing, downloading, and deleting exported meshes. - Images </br> Endpoints for listing, uploading, and deleting images related to avatars. - Create Avatar from images </br> Endpoints to initiate and complete avatar creation from images. - Create Avatar from measurements </br> Endpoint to initiate avatar creation from body measurements. - Create Avatar from scans </br> Endpoints to initiate and complete avatar creation from 3d body scans. - Create Avatar from betas </br> Endpoint to create an avatar from SMPL based beta shape parameters. # Error codes When something goes wrong, the API replies with an additional error code  - `asset_not_found` The requested asset either does not exist, or is not owned by the user (404)  </br> - `too_many_images` The image limit that can be uploaded for avatars from images as been exceeded. (400) </br> - `already_started` A process that already has been requested cannot be started again. (400) </br> - `no_images`  [POST /avatars/create/from-images](#post-/avatars/create/from-images) can only be started with at least one image uploaded  (400) </br> - `inputs_not_ready` Running `/avatars/create/xxxx` endpoint require the inputs to be uploaded (400) </br> - `uuid_invalid_format` Asset ID is in a non-UUID format (400) </br> - ` missing_parameters` Not all required parameters have been supplied for the request (400) </br> - `unauthorized` Trying to access an asset the user does not own, or endpoints that the user is not authorized to call (400) </br> - `too_many_builds` User processing rate has been exceeded. Only one computation heavy process can be started at a time (429) </br> - `asset_not_ready` Cannot call request on an asset which is not in a READY state (400) </br>  # Integration When integrating with our API, you may encounter Cross-Origin Resource Sharing (CORS) issues during deployment. </br> For security reasons, our API does not support direct communication from your frontend. </br> Instead, we recommend that you connect to api.meshcapade through your own backend server. </br> This approach not only mitigates CORS-related challenges but also enhances the overall security of your application. </br> By handling API requests server-side, you can ensure smoother and safer integration. </br> If you encounter any issues during integration, please reach out to us at support@meshcapade.com </br>  # noqa: E501

    The version of the OpenAPI document: v1.20
    Contact: support@meshcapade.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel
from openapi_client.models.docschemas_doc_user_response import DocschemasDocUserResponse

class DocschemasDocUserJSONAPIResponse(BaseModel):
    """
    DocschemasDocUserJSONAPIResponse
    """
    data: Optional[DocschemasDocUserResponse] = None
    __properties = ["data"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> DocschemasDocUserJSONAPIResponse:
        """Create an instance of DocschemasDocUserJSONAPIResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of data
        if self.data:
            _dict['data'] = self.data.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> DocschemasDocUserJSONAPIResponse:
        """Create an instance of DocschemasDocUserJSONAPIResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return DocschemasDocUserJSONAPIResponse.parse_obj(obj)

        _obj = DocschemasDocUserJSONAPIResponse.parse_obj({
            "data": DocschemasDocUserResponse.from_dict(obj.get("data")) if obj.get("data") is not None else None
        })
        return _obj

